#include "LWPack.h"
#include "LWMemory.h"
#include "LWStdLib.h"
#include <string.h>

#undef NIL			/* BEOS has this defined!! */

#define N 4096			/* size of ring buffer */
#define F 18			/* upper limit for MatchSize */
#define THRESHOLD 2		/* encode string into position and length */
						/* if MatchSize is greater than this */
#define NIL N			/* index for root of binary search trees */

static Byte RingBuffer[N+F-1];		/* ring buffer of size N, */
						/* With extra F-1 bytes to facilitate string comparison */

static Word MatchOffset;	/* Offset of string match */
static Word MatchSize;		/* Length of string match 0-18 */
	/* of longest match. These are set by the InsertNode() procedure. */

static Word LeftBranch[N + 1];		/* Left child */
static Word RightBranch[N + 1 + 256];	/* Right child / Hash table */
static Word RootBranch[N + 1];		/* Roots for each binary tree */

 /* left & right children & parents -- These constitute binary search trees. */

/**********************************

	Init the binary tree needed for the compression system

	For i = 0 to N - 1, RightBranch[i] and LeftBranch[i] will be the right and
	left children of node i. These nodes need not be initialized.
	Also, RootBranch[i] is the parent of node i. These are initialized to
	NIL (= N), which stands for 'not used.'
	For i = 0 to 255, RightBranch[N + i + 1] is the root of the tree
	for strings that begin with character i.  These are initialized
	to NIL.  Note there are 256 trees.

**********************************/

static void FASTCALL InitTree(void)	/* initialize trees */
{
	Word i;
	Word *BranchPtr;
	Word *EndPtr;

	i = NIL;

	BranchPtr = &RightBranch[N+1];
	EndPtr = &RightBranch[N+1+256];
	do {
		BranchPtr[0] = i;		/* Zap the hash table */
	} while (++BranchPtr<EndPtr);

	BranchPtr = &RootBranch[0];
	EndPtr = &RootBranch[N];
	do {
		BranchPtr[0] = i;		/* Kill the root indexs */
	} while (++BranchPtr<EndPtr);
}

/**********************************

	Removes a node from the binary tree

**********************************/

static void FASTCALL DeleteNode(Word NodeNum)
{
	Word ClipNum;
	Word Cache,Cache2;

	if (RootBranch[NodeNum] == NIL) {		/* Valid node? */
		return;				/* not in tree */
	}
	if (RightBranch[NodeNum] == NIL) {	/* Is there a right branch? */
		ClipNum = LeftBranch[NodeNum];		/* Clip just the left branch */
	} else if (LeftBranch[NodeNum] == NIL) {	/* Is there a left branch? */
		ClipNum = RightBranch[NodeNum];		/* Clip just the right branch */
	} else {
		ClipNum = LeftBranch[NodeNum];		/* Dual clip tree */
		Cache = RightBranch[ClipNum];
		if (Cache != NIL) {		/* Shall I prune the tree? */
			do {
				ClipNum = Cache;		/* Follow to the end of the tree */
				Cache = RightBranch[ClipNum];
			} while (Cache != NIL);

			Cache = LeftBranch[ClipNum];
			Cache2 = RootBranch[ClipNum];
			RightBranch[Cache2] = Cache;	/* Set the new right parent */
			RootBranch[Cache] = Cache2;	/* Set the new root */

			Cache = LeftBranch[NodeNum];
			LeftBranch[ClipNum] = Cache;		/* Clip the left */
			RootBranch[Cache] = ClipNum;		/* Set the left's new parent */
		}
		Cache = RightBranch[NodeNum];
		RightBranch[ClipNum] = Cache;		/* Attach my tree to the right's end */
		RootBranch[Cache] = ClipNum;		/* Set the new parent */
	}

	/* ClipNum has the branch to remove */

	Cache = RootBranch[NodeNum];		/* Cache a value for speed */

	RootBranch[ClipNum] = Cache;		/* Reset the parent */
	if (RightBranch[Cache] == NodeNum) {	/* Finish up */
		RightBranch[Cache] = ClipNum;		/* Trim the right */
		RootBranch[NodeNum] = NIL;			/* Zap this tree */
		return;
	}
	LeftBranch[Cache] = ClipNum;		/* Trim the left */
	RootBranch[NodeNum] = NIL;			/* Zap this tree */
}


/**********************************

	Inserts string of length F, RingBuffer[r..r+F-1], into one of the
	trees (RingBuffer[r]'th tree) and returns the longest-match position
	and length via the global variables MatchOffset and MatchSize.
	If MatchSize = F, then removes the old node in favor of the new
	one, because the old one will be deleted sooner.
	Note r plays double role, as tree node and position in buffer.

**********************************/

static void FASTCALL InsertNode(Word NodeNum)
{
	Word i;
	Word Parent;	/* Parent node entry */
	int cmp;		/* Branch left or right? (Must be signed!) */
	Byte *key;		/* Pointer to a text string */

	RightBranch[NodeNum] = NIL;	/* Zap the children nodes now */
	LeftBranch[NodeNum] = NIL;

	key = &RingBuffer[NodeNum];
	Parent = (Word)key[0]+(N+1);	/* Get the parent node number via hash table */
	MatchSize = 0;	/* No match yet */
	cmp = 1;		/* Assume a right branch initially */

	for ( ; ; ) {		/* Stay until a match is found */
		Byte *TempPtr;

		if (cmp >= 0) {	/* Right branch? */
			i = RightBranch[Parent];	/* Cache */
			if (i == NIL) {	/* Right branch */
				goto RightEnd;
			}
		} else {
			i = LeftBranch[Parent];	/* Left branch */
			if (i == NIL) {		/* Left branch? */
				goto LeftEnd;
			}
        }

		Parent = i;	/* New parent node */
		i = 1;		/* Init string length */
		TempPtr = &RingBuffer[Parent];
        do {
			cmp = (int)key[i];
			cmp = cmp-(int)TempPtr[i];
            if (cmp)  {		/* Is it different? */
				break;		/* Get out */
			}
		} while (++i<F);
        if (i > MatchSize) {		/* Larger match? */
            MatchOffset = Parent;	/* Set the mark */
			MatchSize = i;			/* Set the length */
            if (i >= F)  {		/* Maximum length! */
				break;		/* Get out of the main loop! */
			}
        }
    }

	/* I have a full match */

    RootBranch[NodeNum] = RootBranch[Parent];	/* Set the new parent */
	LeftBranch[NodeNum] = LeftBranch[Parent];
	RightBranch[NodeNum] = RightBranch[Parent];

    RootBranch[LeftBranch[Parent]] = NodeNum;
	RootBranch[RightBranch[Parent]] = NodeNum;

	i = RootBranch[Parent];
    if (RightBranch[i] == Parent) {
		RightBranch[i] = NodeNum;
	    RootBranch[Parent] = NIL;  /* Remove Parent */
		return;
	}
	LeftBranch[i] = NodeNum;
    RootBranch[Parent] = NIL;  /* Remove Parent */
	return;

LeftEnd:;
	LeftBranch[Parent] = NodeNum;		/* Mark the new left branch */
	RootBranch[NodeNum] = Parent;		/* Mark root */
	return;			/* Exit */

RightEnd:;
	RightBranch[Parent] = NodeNum;		/* Set new right branch */
	RootBranch[NodeNum] = Parent;		/* Mark root */
	return;			/* Exit NOW! */
}

/**********************************

	Compress using the LZSS algorithm
	Return the size in bytes of the compressed data

**********************************/

void ** FASTCALL EncodeLZSS(Byte *InputBuffer,LongWord InputLength)
{
	Word PipedLen;	/* Number of bytes in current string to match */
	Byte BitMask;	/* Bit mask for token flag header (Must be BYTE!) */
	Word i;			/* Temp */
	Word SourceIndex;	/* Index for deleting the nodes */
	Word DestIndex;		/* Index for node insertion */
	Word PrevMatchSize;	/* Size of last run */
	Byte *MaskPtr;		/* Pointer to mask data */
    Word c;

	Byte *OutputBuffer;		/* Pointer to packed data */
	void **OutputHandle;	/* Handle to packed data */

	if (!InputLength) {		/* No input? */
		return 0;			/* Forget it! */
	}

	OutputHandle = AllocAHandle(InputLength+(InputLength>>3)+1);
	if (!OutputHandle) {		/* Can't get memory! */
		return 0;		/* Abort anyway! */
	}
	OutputBuffer = (Byte *)LockAHandle(OutputHandle);	/* Lock it down */
    InitTree();  /* Initialize the binary trees */

	/* Read F bytes into the last F bytes of the buffer */

	PipedLen = N-F;		/* Ring buffer index */
	do {
        RingBuffer[PipedLen] = *InputBuffer++;
		++PipedLen;
		if (!--InputLength) {	/* No more data? */
			break;
		}
	} while (PipedLen<N);
	PipedLen = PipedLen-(N-F);	/* Convert back to count */

	MaskPtr = OutputBuffer;
	MaskPtr[0] = 0;
	++OutputBuffer;
	BitMask = 1;
    SourceIndex = 0;
	DestIndex = N - F;		/* Init the ring buffer index */

	/* Finally, insert the whole string just read.  The */
	/* global variables MatchSize and MatchOffset are set. */

    InsertNode(DestIndex);		/* Insert it in */
    do {

	/* MatchSize may be spuriously long near the end of text. */

	    if (MatchSize > PipedLen) {		/* PipedLen<=F */
			MatchSize = PipedLen;
		}

        if (MatchSize < (THRESHOLD+1)) {

		/* Not long enough match.  Send one byte. */

            MatchSize = 1;		/* Force 1 byte for future loop */
            MaskPtr[0] |= BitMask;  /* 'send one byte' flag */
            *OutputBuffer++ = RingBuffer[DestIndex];  /* Send uncoded. */
        } else {

	/* Send position and length pair. Note MatchSize > THRESHOLD. */

			i = MatchOffset-DestIndex;
			i = i&0xFFF;
			i = i|((MatchSize-(THRESHOLD+1))<<12);
            ((Short *)OutputBuffer)[0] = LoadIntelShort(i);
			OutputBuffer += 2;
        }

		BitMask <<=1;	/* Shift mask left one bit. */
        if (!BitMask) {
			MaskPtr = OutputBuffer;		/* Get the new mask byte */
			++OutputBuffer;			/* Inc past it */
			MaskPtr[0] = 0;			/* Init var */
			BitMask = 1;			/* Init mask */
        }

        PrevMatchSize = MatchSize;	/* Length of the last match */

		i = 0;
		do {
			if (!InputLength) {		/* Any more input? */
		        do {	/* After the end of text, */
		            DeleteNode(SourceIndex);              /* no need to read, but */
					++SourceIndex;
					++DestIndex;
					SourceIndex = SourceIndex & (N - 1);
					DestIndex = DestIndex & (N - 1);
        		    if (PipedLen) {
						--PipedLen;
						InsertNode(DestIndex);		/* buffer may not be empty. */
					}
		        } while (++i<PrevMatchSize);
				break;
			}
			c = *InputBuffer++;		/* Get a byte from the input stream */
			--InputLength;
            DeleteNode(SourceIndex);              /* Delete old strings and */
            RingBuffer[SourceIndex] = c;            /* read new bytes */

	/* If the position is near the end of buffer, extend the buffer to make
		string comparison easier. */

            if (SourceIndex < (F-1)) {
				RingBuffer[SourceIndex + N] = c;
			}
	/* Since this is a ring buffer, increment the position modulo N. */

			++SourceIndex;
			++DestIndex;
            SourceIndex = SourceIndex & (N - 1);
			DestIndex = DestIndex & (N - 1);
            InsertNode(DestIndex);	/* Register the string in RingBuffer[r..r+F-1] */
        } while (++i<PrevMatchSize);
    } while (PipedLen);  /* until length of string to be processed is zero */

	++MaskPtr;
	if (MaskPtr == OutputBuffer) {	/* No data after token? */
		--OutputBuffer;			/* Remove token */
	}

	InputLength = OutputBuffer-(Byte *)(*OutputHandle);
	UnlockAHandle(OutputHandle);		/* Unlock for resizing */
	return ResizeAHandle(OutputHandle,InputLength);	/* Will make smaller */
}

