#include <burger.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define MAXFILENAMELEN 64

#if defined(__IIGS__)
typedef struct {
	LongWord Offset;		/* Offset into the file */
	LongWord Length;		/* Length of the data */
	void **MemPtr;			/* Handle to data in memory */
} MyRezEntry2;				/* Entry in memory */

typedef struct {
	LongWord Type;			/* Data Type */
	LongWord RezNum;		/* Resource number */
	LongWord Count;			/* Number of entries */
	MyRezEntry2 Array[1];	/* First entry */
} MyRezEntry;

typedef struct {
	Byte Name[4];		/* BRGR signature */
	LongWord Count;		/* Number of resource entries */
	LongWord MemSize;	/* Bytes the header will occupy */
} MyRezHeader;
#endif

typedef struct RezEntryInternal {
	struct RezEntryInternal *Next;	/* Forward link */
	struct RezEntryInternal *Prev;	/* Previous link */
	LongWord Length;		/* Length of file */
	LongWord Offset;		/* Offset into dest file */
	LongWord SrcOffset;
	Word Type;				/* Resource type */
	Word RezNum;			/* Resource number */
	char FileName[MAXFILENAMELEN];		/* Filename to copy */
} RezEntryInternal;

typedef struct CacheFile {
	char FileName[MAXFILENAMELEN];	/* Original filename */
	Byte Compress;			/* Compression type */
	Byte Data[1];			/* Data */
} CacheFile;

static LongWord Verbose;		/* True if you want to be briefed on output */
static LongWord EntrySize;		/* Number of resource entries allowed */
static Word LineNum;			/* Line being executed from the script */
static Word RezType;			/* Resource type */
static Word RezIDNum;			/* Resource ID Number */
static Word RezCount;			/* Number of resources loaded */
static Word RezFixed;			/* Fixed memory flag */
static char Delimiters[] = " \t\n";        /* Token delimiters */
static char NumDelimiters[] = " ,\t\n";    /* Value delimiters */
static char InputLine[256];        /* Input line from script */
static char RezFileName[80] = "RezFile";
static LongWord TempOffset;		/* Length of the data output */
static Word LastRezNum;			/* Step value for resources */
static FILE *Headerfp;
static LongWord NoHeaderOverride;	/* If true, then never generate header */
static RezEntryInternal *FirstEntry;
static Word Compress;		/* Is compression enabled? */
static Word Compresstypes[3];	/* Compression types */
static char CacheDirName[] = "CachePck";

#define COMPRESSNONE 0
#define COMPRESSLZSS 1
#define COMPRESSRLE 2
#define COMPRESSLZHUF 3
#define COMPRESSLZARI 4

static char CompressChars[5] = {'.','L','R','H','A'};
static char *CompressNames[5] = {"NONE","LZSS","RLE","LZHUF","LZARI"};

#define COMMANDCOUNT 11      /* Number of commands */
static char *Commands[] = {
	"TYPE","LOAD","ENDIAN","LOADNEXT","LOADFIXED","LOADHANDLE",
	"HEADERFILE","OUTPUTNAME","HLOAD","HLOADNEXT","COMPRESS"
};

#define BUFFER_SIZE 0x40000UL		/* File copy buffer size */
#define ENTRY_SIZE (4000)

static Byte *Buffer;		/* File buffer for data transfer */
static RezEntryInternal *EntArray;	/* Resource headers */
static MyRezHeader MyHeader;		/* Header to save to disk */
static Byte ErrorStr[80];

/**********************************

	Print out a script error

**********************************/

static void PrintError(char *Error)
{
	printf("# Error in Line %d, %s\n",LineNum,Error);
}

/**********************************

	Init the resource file and set up the defaults

**********************************/

static void InitMakeRez(void)
{
	RezType = 1;	/* Default type */
	RezIDNum = 1;	/* Default ID */
	RezCount = 0;	/* No resources loaded */
	RezFixed = 0;  	/* Not fixed memory */
	TempOffset = 0;	/* No files appended */
	FirstEntry = 0;	/* Init the header link */
	Compress = 0;  	/* No compression */
}

/**********************************

	Remove an entry from the entry list using a linked list

**********************************/

static void RemoveEntry(RezEntryInternal *EntryPtr)
{
	if (EntryPtr->Prev) {		/* Previous entry? */
		EntryPtr->Prev->Next = EntryPtr->Next;	/* Remove link */
	} else {
		FirstEntry = EntryPtr->Next;		/* New header link */
	}
	if (EntryPtr->Next) {		/* Next entry? */
		EntryPtr->Next->Prev = EntryPtr->Prev;	/* Set the backward link */
	}
}

/**********************************

	Scan the list of entries and find the lowest number

**********************************/

static RezEntryInternal *FindLowestType(void)
{
	RezEntryInternal *LowestPtr;
	RezEntryInternal *CurrentPtr;
	Word RezNum,Type;

	CurrentPtr = FirstEntry;		/* Init the index */
	if (!CurrentPtr) {			/* No list?? */
		return CurrentPtr;
	}
	RezNum = -1;		/* Use bogus numbers */
	Type = -1;
	do {
		if (CurrentPtr->Type<Type ||		/* Sort by type first */
			(CurrentPtr->Type == Type && CurrentPtr->RezNum<RezNum)) {
			LowestPtr = CurrentPtr;		/* Save as the lowest */
			RezNum = LowestPtr->RezNum;
			Type = LowestPtr->Type;
		}
		CurrentPtr=CurrentPtr->Next;
	} while (CurrentPtr);		/* All scanned? */
	return LowestPtr;
}

/**********************************

	Scan the list of entries and find a specific entry

**********************************/

static RezEntryInternal *FindEntry(Word RezNum,Word Type)
{
	RezEntryInternal *CurrentPtr;

	CurrentPtr = FirstEntry;
	if (CurrentPtr) {
		do {
			if (CurrentPtr->Type==Type && CurrentPtr->RezNum==RezNum) {
				return CurrentPtr;		/* Return the match */
			}
			CurrentPtr = CurrentPtr->Next;
		} while (CurrentPtr);
	}
	return CurrentPtr;
}

/**********************************

	Create runs of structs to sort all the types.
	Store the runs in Buffer and set MyHeader.Count and MemSize
	to reflect the output.

**********************************/

static void CompressHeaders(void)
{
	MyRezEntry *MainPtr;
	if (!FirstEntry) {		/* No resources at all? */
		MyHeader.MemSize = 0;		/* This is easy!! */
		MyHeader.Count = 0;
		return;
	}
	MainPtr = (MyRezEntry *)Buffer;
	do {		/* Perform while resources exist */
		Word Count;
		Word RezNum;
		Word Type;
		MyRezEntry2 *TempPtr;
		RezEntryInternal *EntryPtr;

		++MyHeader.Count;		/* Add one to the count */
		Count = 1;				/* There must be at least 1 */
		EntryPtr = FindLowestType();	/* Find the first entry */
		Type = EntryPtr->Type;
		RezNum = EntryPtr->RezNum;
		MainPtr->Type = Type;		/* Save the type and resource number */
		MainPtr->RezNum = RezNum;
		TempPtr = &MainPtr->Array[0];
		TempPtr->Offset = EntryPtr->Offset;	/* Initial offset and length */
		TempPtr->Length = EntryPtr->Length;
		TempPtr->MemPtr = 0;			/* No memory handle assigned */
		++TempPtr;
		RemoveEntry(EntryPtr);		/* Remove it */
		if (FirstEntry) {
			do {
				++RezNum;
				EntryPtr = FindEntry(RezNum,Type);
				if (!EntryPtr) {
					break;
				}
				TempPtr->Offset = EntryPtr->Offset;
				TempPtr->Length = EntryPtr->Length;
				TempPtr->MemPtr = 0;
				RemoveEntry(EntryPtr);
				++Count;
				++TempPtr;
			} while (FirstEntry);
		}
		MainPtr->Count = Count;		/* Save the count */
		MainPtr = (MyRezEntry *)TempPtr;		/* Point to next buffer */
	} while (FirstEntry);
	MyHeader.MemSize = ((Byte *)MainPtr)-Buffer;
}

/**********************************

	Wrap up the resource file

**********************************/

static void WrapUpMakeRez(void)
{
	FILE *fo;
	LongWord Length;
	Word i;

	if (Verbose) {
		printf("# Now saving data in resource file\n");
	}

	memcpy(MyHeader.Name,"BRGR",4);     /* Init the signature */
	if (Verbose) {
		printf("# Compressing resource header\n");
	}
	CompressHeaders();			/* Compress the resource entry runs */

		/* Adjust all the data offsets to reflect the addition */
		/* of the header and swap endian if needed */

	i = MyHeader.Count;			/* Get the number of runs found */
	if (i) {					/* Any resources? */
		MyRezEntry *MainPtr;

		MainPtr = (MyRezEntry *)Buffer;	/* Index to the buffer */
		Length = MyHeader.MemSize+sizeof(MyHeader);	/* Offset to the end */
		do {
			MyRezEntry2 *TempPtr;
			Word j;

			j = MainPtr->Count;		/* Get the run count */
			#if defined(__BIGENDIAN__)		/* Swap bytes? */
				MainPtr->Count = SwapULong(MainPtr->Count);
				MainPtr->Type = SwapULong(MainPtr->Type);
				MainPtr->RezNum = SwapULong(MainPtr->RezNum);
			#endif
			TempPtr = &MainPtr->Array[0];	/* Index to the array */
			do {
				TempPtr->Offset+=Length;    /* Adjust to true offset */
				#if defined(__BIGENDIAN__)
					TempPtr->Offset = SwapULong(TempPtr->Offset);
					TempPtr->Length = SwapULong(TempPtr->Length);
				#endif
				++TempPtr;		/* Next index */
			} while (--j);
			MainPtr = (MyRezEntry *)TempPtr;	/* Point to end of the table */
		} while (--i);         /* All done? */
	}

	/* Swap endian on the master header */

	Length = MyHeader.MemSize;		/* Get the memory length */
	#if defined(__BIGENDIAN__)
		MyHeader.Count = SwapULong(MyHeader.Count);	/* Swap the header */
		MyHeader.MemSize = SwapULong(Length);
	#endif

	/* Now copy the actual data */

	if (Verbose) {
		printf("# Saving resource header to %s\n",RezFileName);
	}
	fo = OpenAFile(RezFileName,"wb");    /* Open the final file */
	if (!fo) {
		printf("# Can't open the output file %s\n",RezFileName);
		return;
	}
	fwrite(&MyHeader,1,sizeof(MyHeader),fo);     /* Save off the data header */
	fwrite(Buffer,1,Length,fo);
	if (Verbose) {
		printf("# Saving resource data\n");
	}
	i = RezCount;
	if (i) {
		RezEntryInternal *Entry;
		Entry = EntArray;
		do {
			Length = Entry->Length;
			if (Length) {		/* Any data to copy? */
				FILE *fp;
				printf("-");
				fflush(stdout);
				fp = OpenAFile(Entry->FileName,"rb");       /* Reset to the beginning */
				if (fp) {
					fseek(fp,Entry->SrcOffset,SEEK_SET);
					do {
						LongWord Chunk;
						if (Length>BUFFER_SIZE) {	/* Get the smaller, the length */
							Chunk = BUFFER_SIZE;		/* or the buffer size */
						} else {
							Chunk = Length;
						}
						fread(Buffer,1,Chunk,fp);
						fwrite(Buffer,1,Chunk,fo);
						Length -= Chunk;
					} while (Length);
					fclose(fp);              /* Close the temp data file */
				}
			}
			++Entry;
		} while (--i);
	}
	fclose(fo);              /* Close the data file */
	if (Verbose) {
		printf("# Done!!\n");
	}
}

/**********************************

	Create a name for a cached file

**********************************/

static char TempCacheName[40];

static char *MakeCacheName(Word RezNum,Word Type)
{
	sprintf(TempCacheName,"%s:T%u.%u",CacheDirName,RezNum,Type);
	return TempCacheName;
}

/**********************************

	Data compression
	This is really tricky. First I check to see if the resource
	is already in a cache. If so then I use the cached file, otherwise
	I then load and compress the file. If I shrink the file, then I
	place the compressed file into the cache, otherwise I make a copy
	of the file and place it in the cache.

	In all cases, I change the filename to point to the file in the
	cached directory and alter the offset to reflect the fact that it is
	compressed.

	TempOffset has already been added with the uncompressed file
	length, undo this addition if I replace the file with a compressed
	version.

**********************************/

static void HandleCompression(RezEntryInternal *RezPtr)
{
	char OldName[MAXFILENAMELEN];
	FILE *fp;
	Word CompressionUsed;
	void **CompressHandle;
	Word i;
	char *CacheName;
	void *Buffer1,**Buffer2;
	LongWord NewLength;
	TimeDate_t OriginalDate,CacheDate;

	CacheName = MakeCacheName(RezPtr->RezNum,RezPtr->Type);	/* Make the filename */
	if (!GetFileModTime(CacheName,&CacheDate)) {			/* Get cache time */
		if (!GetFileModTime(RezPtr->FileName,&OriginalDate)) {	/* Get original time */
			if (CompareTimeDates(&CacheDate,&OriginalDate)>0) {
				fp = OpenAFile(CacheName,"rb");
				if (fp) {
					NewLength = fgetfilesize(fp);
					if (NewLength>(MAXFILENAMELEN+1)) {	/* Get the filesize */
						fread(&OldName[0],1,MAXFILENAMELEN,fp);
						OldName[MAXFILENAMELEN-1]=0;
						NewLength-=(MAXFILENAMELEN+1);
						CompressionUsed = fgetc(fp);
						if (!stricmp(OldName,RezPtr->FileName)) {
							if (Verbose) {
								printf("# [%s] Using cached copy of %s\n",CompressNames[CompressionUsed],RezPtr->FileName);
							} else {
								putchar(CompressChars[CompressionUsed]);
							}
							fclose(fp);
							if (CompressionUsed==COMPRESSNONE) {
								goto NoPack;		/* Not compressed! */
							}
							goto FixBug;
						}
					}
					fclose(fp);
				}
			}
		}
	}
	if (Verbose) {
		printf("# Compressing %s\n",RezPtr->FileName);
	} else {
		printf("l");
	}

	/* Ok, so I now have to make a cached file... */

	Buffer1 = AllocAPointer(RezPtr->Length);		/* Allocate a buffer for the source file */
	if (!Buffer1) {
		PrintError("Not enough memory for primary pack buffer");
		return;			/* Not enough memory */
	}

	fp = OpenAFile(RezPtr->FileName,"rb");
	if (!fp) {
		PrintError("Can't open input file for packing");
		DeallocAPointer(Buffer1);
		return;
	}
	fread(Buffer1,1,RezPtr->Length,fp);		/* Read in the file */
	fclose(fp);

	/* Let's compress the data */

	NewLength = RezPtr->Length;
	CompressionUsed = COMPRESSNONE;
	CompressHandle = 0;
	i = 0;
	do {
		switch(Compresstypes[i]) {
		case COMPRESSLZSS:
			Buffer2 = EncodeLZSS(Buffer1,RezPtr->Length);	/* Compress the data */
			break;
		case COMPRESSRLE:
			Buffer2 = EncodeRLE(Buffer1,RezPtr->Length);
			break;
		}
		if (!Buffer2) {			/* Didn't get memory? */
			if (Verbose) {
				printf("# Error in compression (%d)\n",Compresstypes[i]);
			}
		} else {
			LongWord TempLen;
			TempLen=GetAHandleSize(Buffer2)+4;
			if (TempLen<NewLength) {
				NewLength = TempLen;
				CompressionUsed = Compresstypes[i];
				CompressHandle = Buffer2;
			} else {
				DeallocAHandle(Buffer2);
			}
		}
	} while (++i<Compress);

	if (CompressionUsed==COMPRESSNONE) {
		DeallocAHandle(CompressHandle);		/* Get rid of the compressed version */
		if (Verbose) {
			printf("# Saving uncompressed copy to cache\n");
		}
		fp = OpenAFile(CacheName,"wb");
		if (fp) {
			fwrite(RezPtr->FileName,1,MAXFILENAMELEN,fp);
			fputc(COMPRESSNONE,fp);
			fwrite(Buffer1,RezPtr->Length,1,fp);
			fclose(fp);
		}
		DeallocAPointer(Buffer1);		/* Release all memory */
		return;			/* Exit */
	}
	DeallocAPointer(Buffer1);		/* Get rid of the original */
	fp = OpenAFile(CacheName,"wb");
	if (fp) {
		if (Verbose) {
			printf("# [%s] Saving compressed data to %s\n",CompressNames[CompressionUsed],CacheName);
		}
		fwrite(RezPtr->FileName,1,MAXFILENAMELEN,fp);
		fputc(CompressionUsed,fp);
		fwritelongl(RezPtr->Length,fp);
		fwrite(LockAHandle(CompressHandle),1,NewLength-4,fp);
		fclose(fp);
	}
	DeallocAHandle(CompressHandle);		/* Release the packed buffer */
FixBug:;
	TempOffset-=(RezPtr->Length-NewLength);	/* Adjust length */
	RezPtr->Length = NewLength;
	i = 0;
	do {
		if (CompressionUsed==Compresstypes[i]) {
			++i;
			break;
		}
	} while (++i<Compress);
	RezPtr->Offset |= i<<29;		/* I am using LZSS */
NoPack:;
	RezPtr->SrcOffset = MAXFILENAMELEN+1;
	strcpy(RezPtr->FileName,CacheName);
}

/**********************************

	Close any open header file

**********************************/

static void CloseHeaderFile(void)
{
	if (Headerfp) {		/* Is there a header being created? */
		if (Verbose) {
			printf("# Closing header file\n");
		}
		fclose(Headerfp);	/* Close the file */
		Headerfp = 0;		/* Mark as closed */
	}
}

/**********************************

	Load in the resource file

	Note : I must increase the resource number even in
	case of an error so that the numbering is not thrown off
	for the resources that DO load.

**********************************/

static void LoadRezFile(Word Config)
{
	char *TextPtr;
	RezEntryInternal *RezPtr;
	FILE *fp;
	LongWord Length;

	RezPtr = &EntArray[RezCount];		/* Get pointer to the entry */
	RezPtr->Type = RezType;				/* Save the new type */
	RezPtr->Offset = TempOffset;		/* Save the new offset */
	RezPtr->SrcOffset = 0;				/* No offset from source */
	RezPtr->Length = 0;					/* No length yet */
	RezPtr->Next = 0;					/* Final link */
	if (RezCount) {					/* Not the first link */
		RezPtr->Prev = RezPtr-1;
		(RezPtr-1)->Next = RezPtr;		/* Insert a forward link */
	} else {
		RezPtr->Prev = 0;
		FirstEntry = EntArray;	/* Init the linked list */
	}
	if (RezFixed) {
		RezPtr->Offset |= 0x80000000;	/* Set the fixed flag */
	}

	/* Should I parse a resource number first (LOAD or HLOAD) */
	/* This will set the proper resource flag */

	if (!(Config&1)) {
		TextPtr = strtok(0,NumDelimiters);      /* Get the operand */
		if (!TextPtr) {
			PrintError("No resource number for LOAD"); /* No operand! */
			return;
		}
		RezPtr->RezNum = atoi(TextPtr);      /* Convert to value */
		LastRezNum = RezPtr->RezNum;
	} else {		/* LOADNEXT or HLOADNEXT */
		++LastRezNum;		/* Increment the resource number */
		RezPtr->RezNum = LastRezNum;	/* Save it */
	}

	/* Should I parse a resource ENUM (HLOAD or HLOADNEXT) */

	if (Config&2) {
		TextPtr = strtok(0,NumDelimiters);
		if (!TextPtr) {
			PrintError("No resource enum for LOAD");
			return;
		}
		if (!NoHeaderOverride) {	/* Can the header be made? */
			if (Headerfp) {
				fprintf(Headerfp,"#define %s %u\n",TextPtr,RezPtr->RezNum);
			}
		}
	}

	/* I must parse a filename */

	TextPtr = strtok(0,NumDelimiters);
	if (!TextPtr) {
		PrintError("No filename for LOAD");
		return;
	}
	if (Verbose) {
		printf("# (%04u) Adding %s to the resource list\n",RezPtr->RezNum,TextPtr);
	}

	fp = OpenAFile(TextPtr,"rb");
	if (!fp) {
		sprintf((char *)ErrorStr,"Can't open %s",TextPtr);
		PrintError((char *)ErrorStr);
		return;
	}

	/* Save off the file information for future parsing */

	Length = fgetfilesize(fp);	/* How large is the input file? */
	fclose(fp);	 				/* Close the temp data file */
	RezPtr->Length = Length;	/* Save the length */
	strcpy(RezPtr->FileName,TextPtr);	/* Copy the filename */
	TempOffset += Length;		/* Add to the offset */
	if (Compress) {
		HandleCompression(RezPtr);	/* Try to compress the file */
	} else {
		if (Verbose) {
			printf("# No compression attempted\n",TextPtr);
		} else {
			printf(".");
		}
	}
	fflush(stdout);
	++RezCount;				/* Valid file loaded in */
}

/**********************************

	Set the Resource type

**********************************/

static void SetType(void)
{
	char *TextPtr;

	TextPtr = strtok(0,Delimiters);     /* Get the operand */
	if (!TextPtr) {
		PrintError("No operand for TYPE");  /* No operand! */
		return;
	}
	RezType = atoi(TextPtr);      /* Convert to value */
	if (Verbose) {
		printf("# Processing resource type %d\n",RezType);
	}
}

/**********************************

	What endian style do I write out my header info?

**********************************/

static void SetEndian(void)
{
	if (Verbose) {
		printf("# Endian is OBSOLETE, ignored\n");
	}
}

/**********************************

	Set the name of the output file

**********************************/

static void SetHeaderFileName(void)
{
	char *TextPtr;

	TextPtr = strtok(0,Delimiters);     /* Get the operand */
	if (!TextPtr) {
		PrintError("No operand for HEADERFILENAME");  /* No operand! */
		return;
	}
	if (Headerfp) {		/* Close off the previous header */
		fclose(Headerfp);
		printf("# Closing header file\n");
	}
	if (!NoHeaderOverride) {	/* Header file allowed? */
		Headerfp = OpenAFile(TextPtr,"w");
		if (!Headerfp) {
			PrintError("Can't open HEADERFILENAME");
			return;
		}
		if (Verbose) {
			printf("# Saving header information in %s\n",TextPtr);
		}
		fprintf(Headerfp,
			"/**********************************\n"
			"\n"
			"\tDefined equates for resource file\n"
			"\n"
			"**********************************/\n\n");
	}

}

/**********************************

	Set the name of the output file

**********************************/

static void SetOutputName(void)
{
	char *TextPtr;

	TextPtr = strtok(0,Delimiters);     /* Get the operand */
	if (!TextPtr) {
		PrintError("No operand for OUTPUTNAME");  /* No operand! */
		return;
	}
	strcpy(RezFileName,TextPtr);      /* Copy the string */
	if (Verbose) {
		printf("# Data will be save to file %s\n",RezFileName);
	}
}

/**********************************

	Set the list of file compressions I can use

**********************************/

static void LogCompressors(void)
{
	char *TextPtr;
	Word i;

	i = 0;
	do {
		TextPtr = strtok(0,NumDelimiters);		/* Get the compressor type */
		if (!TextPtr) {
			if (!i) {
				PrintError("No operand for COMPRESS");
				return;
			}
			break;
		}
		if (!i && !stricmp("OFF",TextPtr)) {
			if (Verbose) {
				printf("Compression is shut off\n");
			}
			Compress=0;
			return;
		}
		if (i) {		/* Check for a comment */
			if (TextPtr[0]==';' || TextPtr[0]=='/') {
				break;
			}
		}
		if (!stricmp("LZSS",TextPtr)) {
			Compresstypes[i] = COMPRESSLZSS;
			continue;
		}
		if (!stricmp("RLE",TextPtr)) {
			Compresstypes[i] = COMPRESSRLE;
			continue;
		}
		PrintError("Only LZSS and RLE are supported");
		return;
	} while (++i<3);
	if (Verbose) {
		printf("# %d compressors have been logged\n",i);
	}
	Compress=i;		/* Save the number of active compressors */
	CreateDirectoryPath(CacheDirName);	/* Create the compression directory */
}

/**********************************

	Execute the script tokens
	Input with the FILE* of the open script file
	I assume LineNum == 0.

**********************************/

static void MakeRez(FILE* fp)
{
	char *TextPtr;  /* Pointer to string token */
	Word i;         /* Index */

	InitMakeRez();    /* Init the output file and other variables */
	while (fgets(InputLine,sizeof(InputLine),fp)) { /* Get a string */
		++LineNum;          /* Adjust the line # */
		TextPtr = strtok(InputLine,Delimiters); /* Get the first token */
		if (!TextPtr) {		/* No good? */
			continue;
		}
		i = 0;      /* Check for the first command */
		if (isalnum(TextPtr[0])) {  /* Comment? */
			do {
				if (!stricmp(TextPtr,Commands[i])) { /* Match? */
					switch (i) {        /* Execute the command */
					case 0:			/* TYPE */
						SetType();	/* Resource type */
						break;
					case 1:			/* LOAD */
						LoadRezFile(0);   /* Input art file type */
						break;
					case 2:			/* ENDIAN */
						SetEndian();    /* Swap endian if needed */
						break;
					case 3:			/* LOADNEXT */
						LoadRezFile(1);	/* Load the next sequential record */
						break;
					case 4:			/* LOADFIXED */
						RezFixed = TRUE;	/* Fixed memory */
						break;
					case 5:			/* LOADHANDLE */
						RezFixed = FALSE;	/* Movable memory */
						break;
					case 6:			/* HEADERFILE */
						SetHeaderFileName();	/* Set the header filename */
						break;
					case 7:			/* OUTPUTNAME */
						SetOutputName();	/* Set the name of the output file */
						break;
					case 8:			/* HLOAD */
						LoadRezFile(2);		/* Get absolute record with name */
						break;
					case 9:			/* HLOADNEXT */
						LoadRezFile(3);		/* Get next record with name */
						break;
					case 10:		/* COMPRESS */
						LogCompressors();	/* Enable data compression */
					}
					break;      /* Don't parse anymore! */
				}
			} while (++i<COMMANDCOUNT); /* Keep checking */
		}
		if (i==COMMANDCOUNT) {      /* Didn't find it? */
			printf("# Command %s not implemented\n",TextPtr);
		}
	}
	CloseHeaderFile();	/* Close any open headers */
	WrapUpMakeRez();	/* Now, I actually make the data file */
}


/**********************************

	Main dispatcher for the converter

**********************************/

static Switch_t SwitchTable[] = {
	{'H',TRUE,&NoHeaderOverride},		/* Do not generate header */
	{'B'|SWITCH_WORD,0,&EntrySize},		/* Number of resources to allocate */
	{'V',TRUE,&Verbose},				/* Verbose mode */
	0
};

int main(int argc,char **argv)
{
	FILE *fp;		/* Input file */
	int Result;		/* Return value */

	NoHeaderOverride = FALSE;		/* Allow saving */
	Verbose = FALSE;				/* Verbose output */
	EntrySize = ENTRY_SIZE;			/* Entry size */
	argc = Switches(argc,argv,SwitchTable);

	if (argc!=2 || !EntrySize) {          /* Gotta have input and output arguments */
		printf(
"# v1.3 Copyright 1997 by LogicWare\n"
"# This program will create a resource data file using a script\n"
"# Usage: MakeRez Infile\n"
"# Switches : [-H] = Do not save the .h file EVER!, [-V] Verbose output\n"
"#   [-B4000] Number of resource entries to allocate internally.\n"
"# Script commands are as follows\n"
"# TYPE 1-65535 [1]\n"
"# ENDIAN BIG | LITTLE [LITTLE]\n"
"# COMPRESS LZSS | NONE [NONE]\n"
"# OUTPUTNAME Filename.rez [RezFile]\n"
"# HEADERFILE Header.h\n"
"# LOADFIXED\n"
"# LOADHANDLE\n"
"# LOAD 1-65535 Filename\n"
"# LOADNEXT Filename\n"
"# HLOAD rRezRef 1-65535 Filename\n"
"# HLOADNEXT rRezRef Filename\n"
			);
		return 1;
	}
	MaxMemSize = 0xC00000;
	InitMemory();				/* Allow the use of handles */
	SetDefaultPrefixs();		/* Init the prefix system */
	SetErrBombFlag(FALSE);

	Result = 1;			/* Assume BAD */
	Buffer = (Byte *)AllocAPointer(BUFFER_SIZE);		/* Allocate a copy buffer */
	if (!Buffer) {
		printf("# Not enough memory for file copy buffer!\n");
	} else {
		if (Verbose) {
			printf("# Allocating %ld bytes for rez entries\n",(LongWord)EntrySize*sizeof(RezEntryInternal));
		}
		EntArray = (RezEntryInternal *)AllocAPointerClear((LongWord)EntrySize*sizeof(RezEntryInternal));
		if (!EntArray) {
			printf("# Not enough memory for resource entries!\n");
		} else {
			fp = fopen(argv[1],"r");    /* Read the ASCII script */
			if (!fp) {
				printf("# Can't open script %s.\n",argv[1]);    /* Oh oh */
			} else {
				if (Verbose) {
					printf("# Processing script %s\n",argv[1]);
				}
				MakeRez(fp);	/* Slice it up! */
				fclose(fp);		/* Close the file */
				Result = 0;		/* No error */
			}
			DeallocAPointer(EntArray);	/* Release the memory */
		}
		DeallocAPointer(Buffer);		/* Release the first pointer */
	}
	return Result;		/* Exit */
}
